"""Blob file storage

Contains blob file storage interfaces. See `on_disk.py` for an example of a concrete implementation.

BlobStorageClient, BlobStorageClientConfig and BlobStorageConnection
-----------------------------------------------------

```
blob_storage_client_cls: Type[BlobStorageClient]
blob_storage_client_config: BlobStorageClientConfig

blob_storage_client = blob_storage_client_cls(config=blob_storage_client_config)
```

BlobStorageClient implements context manager (`__enter__()` and `__exit__()`) to create a BlobStorageConnection.
`BlobStorageClient.__enter__()` creates a connection (BlobStorageConnection) to the file system (e.g. SeaweedFS).
BlobStorageConnection implements operations on files (read, write, ...).

```
with blob_storage_client as conn:
    conn.read(...)
```

See `blob_storage/service.py` for usage example.

Write/persist SyftObject to blob storage
----------------------------------------

- create a CreateBlobStorageEntry from SyftObject `create_blob_storage_entry = CreateBlobStorageEntry.from(obj)`
- pre-allocate the file object `blob_deposit = api.services.blob_storage.allocate(create_blob_storage_entry)`
  (this returns a BlobDeposit)
- use `BlobDeposit.write` to upload/save/persist the SyftObject
  `blob_deposit.write(sy.serialize(user_object, to_bytes=True))`

Read/retrieve SyftObject from blob storage
------------------------------------------

- get a BlobRetrieval from the id of the BlobStorageEntry of the SyftObject
  `blob_retrieval = api.services.blob_storage.read(blob_storage_entry_id)`
- use `BlobRetrieval.read` to retrieve the SyftObject `syft_object = blob_retrieval.read()`
"""

# stdlib
from collections.abc import Callable
from collections.abc import Generator
from io import BytesIO
import logging
from typing import Any

# third party
from pydantic import BaseModel
import requests
from typing_extensions import Self

# relative
from ...serde.deserialize import _deserialize as deserialize
from ...serde.serializable import serializable
from ...service.response import SyftError
from ...service.response import SyftSuccess
from ...types.base import SyftBaseModel
from ...types.blob_storage import BlobFile
from ...types.blob_storage import BlobFileType
from ...types.blob_storage import BlobStorageEntry
from ...types.blob_storage import CreateBlobStorageEntry
from ...types.blob_storage import DEFAULT_CHUNK_SIZE
from ...types.blob_storage import SecureFilePathLocation
from ...types.server_url import ServerURL
from ...types.syft_migration import migrate
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import drop
from ...types.transforms import make_set_default
from ...types.uid import UID

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10
MAX_RETRIES = 20


@serializable()
class BlobRetrieval(SyftObject):
    __canonical_name__ = "BlobRetrieval"
    __version__ = SYFT_OBJECT_VERSION_1

    type_: type | None = None
    file_name: str
    syft_blob_storage_entry_id: UID | None = None
    file_size: int | None = None


@serializable()
class SyftObjectRetrieval(BlobRetrieval):
    __canonical_name__ = "SyftObjectRetrieval"
    __version__ = SYFT_OBJECT_VERSION_1

    syft_object: bytes

    def _read_data(
        self, stream: bool = False, _deserialize: bool = True, **kwargs: Any
    ) -> Any:
        # development setup, we can access the same filesystem
        if not _deserialize:
            res = self.syft_object
        else:
            res = deserialize(self.syft_object, from_bytes=True)

        # TODO: implement proper streaming from local files
        if stream:
            return [res]
        else:
            return res

    def read(self, _deserialize: bool = True) -> SyftObject | SyftError:
        return self._read_data(_deserialize=_deserialize)


def syft_iter_content(
    blob_url: str | ServerURL,
    chunk_size: int,
    max_retries: int = MAX_RETRIES,
    timeout: int = DEFAULT_TIMEOUT,
) -> Generator:
    """Custom iter content with smart retries (start from last byte read)"""
    current_byte = 0
    for attempt in range(max_retries):
        headers = {"Range": f"bytes={current_byte}-"}
        try:
            with requests.get(
                str(blob_url), stream=True, headers=headers, timeout=(timeout, timeout)
            ) as response:
                response.raise_for_status()
                for chunk in response.iter_content(
                    chunk_size=chunk_size, decode_unicode=False
                ):
                    current_byte += len(chunk)
                    yield chunk
            return  # If successful, exit the function
        except requests.exceptions.RequestException as e:
            if attempt < max_retries:
                logger.debug(
                    f"Attempt {attempt}/{max_retries} failed: {e} at byte {current_byte}. Retrying..."
                )
            else:
                logger.error(f"Max retries reached - {e}")
                raise


@serializable()
class BlobRetrievalByURL(BlobRetrieval):
    __canonical_name__ = "BlobRetrievalByURL"
    __version__ = SYFT_OBJECT_VERSION_1

    url: ServerURL | str
    proxy_server_uid: UID | None = None

    def read(self) -> SyftObject | SyftError:
        if self.type_ is BlobFileType:
            return BlobFile(
                file_name=self.file_name,
                syft_client_verify_key=self.syft_client_verify_key,
                syft_server_location=self.syft_server_location,
                syft_blob_storage_entry_id=self.syft_blob_storage_entry_id,
                file_size=self.file_size,
            )
        else:
            return self._read_data()

    def _read_data(
        self,
        stream: bool = False,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        # relative
        from ...client.api import APIRegistry

        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )

        if api and api.connection and isinstance(self.url, ServerURL):
            if self.proxy_server_uid is None:
                blob_url = api.connection.to_blob_route(
                    self.url.url_path, host=self.url.host_or_ip
                )
            else:
                blob_url = api.connection.stream_via(
                    self.proxy_server_uid, self.url.url_path
                )
                stream = True
        else:
            blob_url = self.url

        try:
            is_blob_file = self.type_ is not None and issubclass(
                self.type_, BlobFileType
            )
            if is_blob_file and stream:
                return syft_iter_content(blob_url, chunk_size)

            response = requests.get(str(blob_url), stream=stream)  # nosec
            resp_content = response.content
            response.raise_for_status()

            return (
                resp_content
                if is_blob_file
                else deserialize(resp_content, from_bytes=True)
            )
        except requests.RequestException as e:
            return SyftError(message=f"Failed to retrieve with error: {e}")


@serializable()
class BlobDeposit(SyftObject):
    __canonical_name__ = "BlobDeposit"
    __version__ = SYFT_OBJECT_VERSION_1

    blob_storage_entry_id: UID

    def write(self, data: BytesIO) -> SyftSuccess | SyftError:
        raise NotImplementedError


@serializable(canonical_name="BlobStorageClientConfig", version=1)
class BlobStorageClientConfig(BaseModel):
    pass


class BlobStorageConnection:
    def __enter__(self) -> Self:
        raise NotImplementedError

    def __exit__(self, *exc: Any) -> None:
        raise NotImplementedError

    def read(self, fp: SecureFilePathLocation, type_: type | None) -> BlobRetrieval:
        raise NotImplementedError

    def allocate(
        self, obj: CreateBlobStorageEntry
    ) -> SecureFilePathLocation | SyftError:
        raise NotImplementedError

    def write(self, obj: BlobStorageEntry) -> BlobDeposit:
        raise NotImplementedError

    def delete(self, fp: SecureFilePathLocation) -> bool:
        raise NotImplementedError


@serializable(canonical_name="BlobStorageClient", version=1)
class BlobStorageClient(SyftBaseModel):
    config: BlobStorageClientConfig

    def connect(self) -> BlobStorageConnection:
        raise NotImplementedError


@serializable(canonical_name="BlobStorageConfig", version=1)
class BlobStorageConfig(SyftBaseModel):
    client_type: type[BlobStorageClient]
    client_config: BlobStorageClientConfig
    min_blob_size: int = 0  # in MB



# future
from __future__ import annotations

# stdlib
from collections.abc import Callable
import types
import typing
from typing import Any

# third party
from pydantic import BaseModel
from pydantic import Field
from result import Err
from result import Ok
from result import Result
from typeguard import check_type

# relative
from ..serde.serializable import serializable
from ..server.credentials import SyftSigningKey
from ..server.credentials import SyftVerifyKey
from ..service.action.action_permissions import ActionObjectPermission
from ..service.action.action_permissions import StoragePermission
from ..service.context import AuthedServiceContext
from ..service.response import SyftSuccess
from ..types.base import SyftBaseModel
from ..types.syft_object import BaseDateTime
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.uid import UID
from ..util.telemetry import instrument
from .locks import LockingConfig
from .locks import NoLockingConfig
from .locks import SyftLock


@serializable(canonical_name="BasePartitionSettings", version=1)
class BasePartitionSettings(SyftBaseModel):
    """Basic Partition Settings

    Parameters:
        name: str
            Identifier to be used as prefix by stores and for partitioning
    """

    name: str


def first_or_none(result: Any) -> Ok:
    if hasattr(result, "__len__") and len(result) > 0:
        return Ok(result[0])
    return Ok(None)


def is_generic_alias(t: type) -> bool:
    return isinstance(t, types.GenericAlias | typing._GenericAlias)


class StoreClientConfig(BaseModel):
    """Base Client specific configuration"""

    pass


@serializable(canonical_name="PartitionKey", version=1)
class PartitionKey(BaseModel):
    key: str
    type_: type | object

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) == type(self)
            and self.key == other.key
            and self.type_ == other.type_
        )

    def with_obj(self, obj: Any) -> QueryKey:
        return QueryKey.from_obj(partition_key=self, obj=obj)

    def extract_list(self, obj: Any) -> list:
        # not a list and matches the internal list type of the _GenericAlias
        if not isinstance(obj, list):
            if not isinstance(obj, typing.get_args(self.type_)):
                obj = getattr(obj, self.key)
                if isinstance(obj, types.FunctionType | types.MethodType):
                    obj = obj()

            if not isinstance(obj, list) and isinstance(
                obj, typing.get_args(self.type_)
            ):
                # still not a list but the right type
                obj = [obj]

        # is a list type so lets compare directly
        check_type(obj, self.type_)
        return obj

    @property
    def type_list(self) -> bool:
        return is_generic_alias(self.type_) and self.type_.__origin__ == list


@serializable(canonical_name="PartitionKeys", version=1)
class PartitionKeys(BaseModel):
    pks: PartitionKey | tuple[PartitionKey, ...] | list[PartitionKey]

    @property
    def all(self) -> tuple[PartitionKey, ...] | list[PartitionKey]:
        # make sure we always return a list even if there's a single value
        return self.pks if isinstance(self.pks, tuple | list) else [self.pks]

    def with_obj(self, obj: Any) -> QueryKeys:
        return QueryKeys.from_obj(partition_keys=self, obj=obj)

    def with_tuple(self, *args: Any) -> QueryKeys:
        return QueryKeys.from_tuple(partition_keys=self, args=args)

    def add(self, pk: PartitionKey) -> PartitionKeys:
        return PartitionKeys(pks=list(self.all) + [pk])

    @staticmethod
    def from_dict(cks_dict: dict[str, type]) -> PartitionKeys:
        pks = []
        for k, t in cks_dict.items():
            pks.append(PartitionKey(key=k, type_=t))
        return PartitionKeys(pks=pks)


@serializable(canonical_name="QueryKey", version=1)
class QueryKey(PartitionKey):
    value: Any = None

    def __eq__(self, other: Any) -> bool:
        return (
            type(other) == type(self)
            and self.key == other.key
            and self.type_ == other.type_
            and self.value == other.value
        )

    @property
    def partition_key(self) -> PartitionKey:
        return PartitionKey(key=self.key, type_=self.type_)

    @staticmethod
    def from_obj(partition_key: PartitionKey, obj: Any) -> QueryKey:
        pk_key = partition_key.key
        pk_type = partition_key.type_

        # 🟡 TODO: support more advanced types than List[type]
        if partition_key.type_list:
            pk_value = partition_key.extract_list(obj)
        else:
            if isinstance(obj, pk_type):
                pk_value = obj
            else:
                pk_value = getattr(obj, pk_key)
                # object has a method for getting these types
                # we can't use properties because we don't seem to be able to get the
                # return types
                # TODO: fix the mypy issue
                if isinstance(pk_value, types.FunctionType | types.MethodType):  # type: ignore[unreachable]
                    pk_value = pk_value()  # type: ignore[unreachable]

            if pk_value and not isinstance(pk_value, pk_type):
                raise Exception(
                    f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                )
        return QueryKey(key=pk_key, type_=pk_type, value=pk_value)

    @property
    def as_dict(self) -> dict[str, Any]:
        return {self.key: self.value}

    @property
    def as_dict_mongo(self) -> dict[str, Any]:
        key = self.key
        if key == "id":
            key = "_id"
        if self.type_list:
            # We want to search inside the list of values
            return {key: {"$in": self.value}}
        return {key: self.value}


@serializable(canonical_name="PartitionKeysWithUID", version=1)
class PartitionKeysWithUID(PartitionKeys):
    uid_pk: PartitionKey

    @property
    def all(self) -> tuple[PartitionKey, ...] | list[PartitionKey]:
        all_keys = list(self.pks) if isinstance(self.pks, tuple | list) else [self.pks]
        if self.uid_pk not in all_keys:
            all_keys.insert(0, self.uid_pk)
        return all_keys


@serializable(canonical_name="QueryKeys", version=1)
class QueryKeys(SyftBaseModel):
    qks: QueryKey | tuple[QueryKey, ...] | list[QueryKey]

    @property
    def all(self) -> tuple[QueryKey, ...] | list[QueryKey]:
        # make sure we always return a list even if there's a single value
        return self.qks if isinstance(self.qks, tuple | list) else [self.qks]

    @staticmethod
    def from_obj(partition_keys: PartitionKeys, obj: SyftObject) -> QueryKeys:
        qks = []
        for partition_key in partition_keys.all:
            pk_key = partition_key.key  # name of the attribute
            pk_type = partition_key.type_
            pk_value = getattr(obj, pk_key)
            # object has a method for getting these types
            # we can't use properties because we don't seem to be able to get the
            # return types
            if isinstance(pk_value, types.FunctionType | types.MethodType):
                pk_value = pk_value()
            if partition_key.type_list:
                pk_value = partition_key.extract_list(obj)
            else:
                if pk_value and not isinstance(pk_value, pk_type):
                    raise Exception(
                        f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                    )
            qk = QueryKey(key=pk_key, type_=pk_type, value=pk_value)
            qks.append(qk)
        return QueryKeys(qks=qks)

    @staticmethod
    def from_tuple(partition_keys: PartitionKeys, args: tuple) -> QueryKeys:
        qks = []
        for partition_key, pk_value in zip(partition_keys.all, args):
            pk_key = partition_key.key
            pk_type = partition_key.type_
            if not isinstance(pk_value, pk_type):
                raise Exception(
                    f"PartitionKey {pk_value} of type {type(pk_value)} must be {pk_type}."
                )
            qk = QueryKey(key=pk_key, type_=pk_type, value=pk_value)
            qks.append(qk)
        return QueryKeys(qks=qks)

    @staticmethod
    def from_dict(qks_dict: dict[str, Any]) -> QueryKeys:
        qks = []
        for k, v in qks_dict.items():
            qks.append(QueryKey(key=k, type_=type(v), value=v))
        return QueryKeys(qks=qks)

    @property
    def as_dict(self) -> dict:
        qk_dict = {}
        for qk in self.all:
            qk_key = qk.key
            qk_value = qk.value
            qk_dict[qk_key] = qk_value
        return qk_dict

    @property
    def as_dict_mongo(self) -> dict:
        qk_dict = {}
        for qk in self.all:
            qk_key = qk.key
            qk_value = qk.value
            if qk_key == "id":
                qk_key = "_id"
            if qk.type_list:
                # We want to search inside the list of values
                qk_dict[qk_key] = {"$in": qk_value}
            else:
                qk_dict[qk_key] = qk_value
        return qk_dict


UIDPartitionKey = PartitionKey(key="id", type_=UID)


@serializable(canonical_name="PartitionSettings", version=1)
class PartitionSettings(BasePartitionSettings):
    object_type: type
    store_key: PartitionKey = UIDPartitionKey

    @property
    def unique_keys(self) -> PartitionKeys:
        unique_keys = PartitionKeys.from_dict(self.object_type._syft_unique_keys_dict())
        return unique_keys.add(self.store_key)

    @property
    def searchable_keys(self) -> PartitionKeys:
        return PartitionKeys.from_dict(self.object_type._syft_searchable_keys_dict())


@instrument
@serializable(
    attrs=["settings", "store_config", "unique_cks", "searchable_cks"],
    canonical_name="StorePartition",
    version=1,
)
class StorePartition:
    """Base StorePartition

    Parameters:
        settings: PartitionSettings
            PySyft specific settings
        store_config: StoreConfig
            Backend specific configuration
    """

    def __init__(
        self,
        server_uid: UID,
        root_verify_key: SyftVerifyKey | None,
        settings: PartitionSettings,
        store_config: StoreConfig,
        has_admin_permissions: Callable[[SyftVerifyKey], bool] | None = None,
    ) -> None:
        if root_verify_key is None:
            root_verify_key = SyftSigningKey.generate().verify_key
        self.server_uid = server_uid
        self.root_verify_key = root_verify_key
        self.settings = settings
        self.store_config = store_config
        self.has_admin_permissions = has_admin_permissions
        res = self.init_store()
        if res.is_err():
            raise RuntimeError(
                f"Something went wrong initializing the store: {res.err()}"
            )

        store_config.locking_config.lock_name = f"StorePartition-{settings.name}"
        self.lock = SyftLock(store_config.locking_config)

    def init_store(self) -> Result[Ok, Err]:
        try:
            self.unique_cks = self.settings.unique_keys.all
            self.searchable_cks = self.settings.searchable_keys.all
        except BaseException as e:
            return Err(str(e))

        return Ok(True)

    def matches_unique_cks(self, partition_key: PartitionKey) -> bool:
        return partition_key in self.unique_cks

    def matches_searchable_cks(self, partition_key: PartitionKey) -> bool:
        return partition_key in self.searchable_cks

    def store_query_key(self, obj: Any) -> QueryKey:
        return self.settings.store_key.with_obj(obj)

    def store_query_keys(self, objs: Any) -> QueryKeys:
        return QueryKeys(qks=[self.store_query_key(obj) for obj in objs])

    # Thread-safe methods
    def _thread_safe_cbk(self, cbk: Callable, *args: Any, **kwargs: Any) -> Any | Err:
        locked = self.lock.acquire(blocking=True)
        if not locked:
            return Err(
                f"Failed to acquire lock for the operation {self.lock.lock_name} ({self.lock._lock})"
            )

        try:
            result = cbk(*args, **kwargs)
        except BaseException as e:
            result = Err(str(e))
        self.lock.release()

        return result

    def set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftObject,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        if obj.created_date is None:
            obj.created_date = BaseDateTime.now()
        return self._thread_safe_cbk(
            self._set,
            credentials=credentials,
            obj=obj,
            add_permissions=add_permissions,
            add_storage_permission=add_storage_permission,
            ignore_duplicates=ignore_duplicates,
        )

    def get(
        self,
        credentials: SyftVerifyKey,
        uid: UID,
    ) -> Result[SyftObject, str]:
        return self._thread_safe_cbk(
            self._get,
            uid=uid,
            credentials=credentials,
        )

    def find_index_or_search_keys(
        self,
        credentials: SyftVerifyKey,
        index_qks: QueryKeys,
        search_qks: QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[SyftObject], str]:
        return self._thread_safe_cbk(
            self._find_index_or_search_keys,
            credentials,
            index_qks=index_qks,
            search_qks=search_qks,
            order_by=order_by,
        )

    def remove_keys(
        self,
        unique_query_keys: QueryKeys,
        searchable_query_keys: QueryKeys,
    ) -> None:
        self._thread_safe_cbk(
            self._remove_keys,
            unique_query_keys=unique_query_keys,
            searchable_query_keys=searchable_query_keys,
        )

    def update(
        self,
        credentials: SyftVerifyKey,
        qk: QueryKey,
        obj: SyftObject,
        has_permission: bool = False,
    ) -> Result[SyftObject, str]:
        return self._thread_safe_cbk(
            self._update,
            credentials=credentials,
            qk=qk,
            obj=obj,
            has_permission=has_permission,
        )

    def get_all_from_store(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[SyftObject], str]:
        return self._thread_safe_cbk(
            self._get_all_from_store, credentials, qks, order_by
        )

    def delete(
        self, credentials: SyftVerifyKey, qk: QueryKey, has_permission: bool = False
    ) -> Result[SyftSuccess, Err]:
        return self._thread_safe_cbk(
            self._delete, credentials, qk, has_permission=has_permission
        )

    def all(
        self,
        credentials: SyftVerifyKey,
        order_by: PartitionKey | None = None,
        has_permission: bool | None = False,
    ) -> Result[list[BaseStash.object_type], str]:
        return self._thread_safe_cbk(self._all, credentials, order_by, has_permission)

    def migrate_data(
        self,
        to_klass: SyftObject,
        context: AuthedServiceContext,
        has_permission: bool | None = False,
    ) -> Result[bool, str]:
        return self._thread_safe_cbk(
            self._migrate_data, to_klass, context, has_permission
        )

    # Potentially thread-unsafe methods.
    # CAUTION:
    #       * Don't use self.lock here.
    #       * Do not call the public thread-safe methods here(with locking).
    # These methods are called from the public thread-safe API, and will hang the process.
    def _set(
        self,
        credentials: SyftVerifyKey,
        obj: SyftObject,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[SyftObject, str]:
        raise NotImplementedError

    def _update(
        self,
        credentials: SyftVerifyKey,
        qk: QueryKey,
        obj: SyftObject,
        has_permission: bool = False,
        overwrite: bool = False,
        allow_missing_keys: bool = False,
    ) -> Result[SyftObject, str]:
        raise NotImplementedError

    def _get_all_from_store(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[SyftObject], str]:
        raise NotImplementedError

    def _delete(
        self, credentials: SyftVerifyKey, qk: QueryKey, has_permission: bool = False
    ) -> Result[SyftSuccess, Err]:
        raise NotImplementedError

    def _all(
        self,
        credentials: SyftVerifyKey,
        order_by: PartitionKey | None = None,
        has_permission: bool | None = False,
    ) -> Result[list[BaseStash.object_type], str]:
        raise NotImplementedError

    def add_permission(self, permission: ActionObjectPermission) -> None:
        raise NotImplementedError

    def add_permissions(self, permissions: list[ActionObjectPermission]) -> None:
        raise NotImplementedError

    def remove_permission(self, permission: ActionObjectPermission) -> None:
        raise NotImplementedError

    def has_permission(self, permission: ActionObjectPermission) -> bool:
        raise NotImplementedError

    def get_all_permissions(self) -> Result[dict[UID, set[str]], str]:
        raise NotImplementedError

    def _get_permissions_for_uid(self, uid: UID) -> Result[set[str], str]:
        raise NotImplementedError

    def add_storage_permission(self, permission: StoragePermission) -> None:
        raise NotImplementedError

    def add_storage_permissions(self, permissions: list[StoragePermission]) -> None:
        raise NotImplementedError

    def remove_storage_permission(self, permission: StoragePermission) -> None:
        raise NotImplementedError

    def has_storage_permission(self, permission: StoragePermission | UID) -> bool:
        raise NotImplementedError

    def _get_storage_permissions_for_uid(self, uid: UID) -> Result[set[UID], str]:
        raise NotImplementedError

    def get_all_storage_permissions(self) -> Result[dict[UID, set[UID]], str]:
        raise NotImplementedError

    def _migrate_data(
        self,
        to_klass: SyftObject,
        context: AuthedServiceContext,
        has_permission: bool,
    ) -> Result[bool, str]:
        raise NotImplementedError


@instrument
@serializable(canonical_name="DocumentStore", version=1)
class DocumentStore:
    """Base Document Store

    Parameters:
        store_config: StoreConfig
            Store specific configuration.
    """

    partitions: dict[str, StorePartition]
    partition_type: type[StorePartition]

    def __init__(
        self,
        server_uid: UID,
        root_verify_key: SyftVerifyKey | None,
        store_config: StoreConfig,
    ) -> None:
        if store_config is None:
            raise Exception("must have store config")
        self.partitions = {}
        self.store_config = store_config
        self.server_uid = server_uid
        self.root_verify_key = root_verify_key

    def __has_admin_permissions(
        self, settings: PartitionSettings
    ) -> Callable[[SyftVerifyKey], bool]:
        # relative
        from ..service.user.user import User
        from ..service.user.user_roles import ServiceRole
        from ..service.user.user_stash import UserStash

        # leave out UserStash to avoid recursion
        # TODO: pass the callback from BaseStash instead of DocumentStore
        # so that this works with UserStash after the sqlite thread fix is merged
        if settings.object_type is User:
            return lambda credentials: False

        user_stash = UserStash(store=self)

        def has_admin_permissions(credentials: SyftVerifyKey) -> bool:
            res = user_stash.get_by_verify_key(
                credentials=credentials,
                verify_key=credentials,
            )

            return (
                res.is_ok()
                and (user := res.ok()) is not None
                and user.role in (ServiceRole.DATA_OWNER, ServiceRole.ADMIN)
            )

        return has_admin_permissions

    def partition(self, settings: PartitionSettings) -> StorePartition:
        if settings.name not in self.partitions:
            self.partitions[settings.name] = self.partition_type(
                server_uid=self.server_uid,
                root_verify_key=self.root_verify_key,
                settings=settings,
                store_config=self.store_config,
                has_admin_permissions=self.__has_admin_permissions(settings),
            )
        return self.partitions[settings.name]

    def get_partition_object_types(self) -> list[type]:
        return [
            partition.settings.object_type for partition in self.partitions.values()
        ]


@instrument
class BaseStash:
    object_type: type[SyftObject]
    settings: PartitionSettings
    partition: StorePartition

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.partition = store.partition(type(self).settings)

    def check_type(self, obj: Any, type_: type) -> Result[Any, str]:
        return (
            Ok(obj)
            if isinstance(obj, type_)
            else Err(f"{type(obj)} does not match required type: {type_}")
        )

    def get_all(
        self,
        credentials: SyftVerifyKey,
        order_by: PartitionKey | None = None,
        has_permission: bool = False,
    ) -> Result[list[BaseStash.object_type], str]:
        return self.partition.all(credentials, order_by, has_permission)

    def add_permissions(self, permissions: list[ActionObjectPermission]) -> None:
        self.partition.add_permissions(permissions)

    def add_permission(self, permission: ActionObjectPermission) -> None:
        self.partition.add_permission(permission)

    def remove_permission(self, permission: ActionObjectPermission) -> None:
        self.partition.remove_permission(permission)

    def has_permission(self, permission: ActionObjectPermission) -> bool:
        return self.partition.has_permission(permission=permission)

    def has_storage_permission(self, permission: StoragePermission) -> bool:
        return self.partition.has_storage_permission(permission=permission)

    def __len__(self) -> int:
        return len(self.partition)

    def set(
        self,
        credentials: SyftVerifyKey,
        obj: BaseStash.object_type,
        add_permissions: list[ActionObjectPermission] | None = None,
        add_storage_permission: bool = True,
        ignore_duplicates: bool = False,
    ) -> Result[BaseStash.object_type, str]:
        res = self.check_type(obj, self.object_type)
        # we dont use and_then logic here as it is hard because of the order of the arguments
        if res.is_err():
            return res
        res = self.partition.set(
            credentials=credentials,
            obj=obj,
            ignore_duplicates=ignore_duplicates,
            add_permissions=add_permissions,
            add_storage_permission=add_storage_permission,
        )

        return res

    def query_all(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKey | QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[list[BaseStash.object_type], str]:
        if isinstance(qks, QueryKey):
            qks = QueryKeys(qks=qks)

        unique_keys = []
        searchable_keys = []

        for qk in qks.all:
            pk = qk.partition_key
            if self.partition.matches_unique_cks(pk):
                unique_keys.append(qk)
            elif self.partition.matches_searchable_cks(pk):
                searchable_keys.append(qk)
            else:
                return Err(
                    f"{qk} not in {type(self.partition)} unique or searchable keys"
                )

        index_qks = QueryKeys(qks=unique_keys)
        search_qks = QueryKeys(qks=searchable_keys)

        return self.partition.find_index_or_search_keys(
            credentials=credentials,
            index_qks=index_qks,
            search_qks=search_qks,
            order_by=order_by,
        )

    def query_all_kwargs(
        self,
        credentials: SyftVerifyKey,
        **kwargs: dict[str, Any],
    ) -> Result[list[BaseStash.object_type], str]:
        order_by = kwargs.pop("order_by", None)
        _ = kwargs.pop("created_date", None)
        _ = kwargs.pop("deleted_date", None)
        _ = kwargs.pop("updated_date", None)
        qks = QueryKeys.from_dict(kwargs)
        return self.query_all(credentials=credentials, qks=qks, order_by=order_by)

    def query_one(
        self,
        credentials: SyftVerifyKey,
        qks: QueryKey | QueryKeys,
        order_by: PartitionKey | None = None,
    ) -> Result[BaseStash.object_type | None, str]:
        return self.query_all(
            credentials=credentials, qks=qks, order_by=order_by
        ).and_then(first_or_none)

    def query_one_kwargs(
        self,
        credentials: SyftVerifyKey,
        **kwargs: dict[str, Any],
    ) -> Result[BaseStash.object_type | None, str]:
        return self.query_all_kwargs(credentials, **kwargs).and_then(first_or_none)

    def find_all(
        self, credentials: SyftVerifyKey, **kwargs: dict[str, Any]
    ) -> Result[list[BaseStash.object_type], str]:
        return self.query_all_kwargs(credentials=credentials, **kwargs)

    def find_one(
        self, credentials: SyftVerifyKey, **kwargs: dict[str, Any]
    ) -> Result[BaseStash.object_type | None, str]:
        return self.query_one_kwargs(credentials=credentials, **kwargs)

    def find_and_delete(
        self, credentials: SyftVerifyKey, **kwargs: dict[str, Any]
    ) -> Result[SyftSuccess, Err]:
        obj = self.query_one_kwargs(credentials=credentials, **kwargs)
        if obj.is_err():
            return obj
        else:
            obj = obj.ok()

        if not obj:
            return Err(f"Object does not exists with kwargs: {kwargs}")
        qk = self.partition.store_query_key(obj)
        return self.delete(credentials=credentials, qk=qk)

    def delete(
        self, credentials: SyftVerifyKey, qk: QueryKey, has_permission: bool = False
    ) -> Result[SyftSuccess, Err]:
        return self.partition.delete(
            credentials=credentials, qk=qk, has_permission=has_permission
        )

    def update(
        self,
        credentials: SyftVerifyKey,
        obj: BaseStash.object_type,
        has_permission: bool = False,
    ) -> Result[BaseStash.object_type, str]:
        obj.updated_date = BaseDateTime.now()
        qk = self.partition.store_query_key(obj)
        res = self.partition.update(
            credentials=credentials, qk=qk, obj=obj, has_permission=has_permission
        )
        return res

    def delete_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[SyftSuccess, str]:
        qk = UIDPartitionKey.with_obj(uid)
        result = self.delete(credentials=credentials, qk=qk)
        if result.is_ok():
            return Ok(SyftSuccess(message=f"ID: {uid} deleted"))
        return result

    def get_by_uid(
        self, credentials: SyftVerifyKey, uid: UID
    ) -> Result[BaseStash.object_type | None, str]:
        res = self.partition.get(credentials=credentials, uid=uid)

        # NOTE Return Ok(None) when no results are found for backwards compatibility
        if res.is_err():
            return Ok(None)
        return res


@instrument
class BaseUIDStoreStash(BaseStash):
    pass


@serializable()
class StoreConfig(SyftBaseObject):
    """Base Store configuration

    Parameters:
        store_type: Type
            Document Store type
        client_config: Optional[StoreClientConfig]
            Backend-specific config
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to NoLockingConfig.
    """

    __canonical_name__ = "StoreConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    store_type: type[DocumentStore]
    client_config: StoreClientConfig | None = None
    locking_config: LockingConfig = Field(default_factory=NoLockingConfig)



# stdlib
import json
import logging
import secrets
from typing import Any

# third party
from IPython.display import HTML
from IPython.display import display
import jinja2

# relative
from ....types.uid import UID
from ...assets import load_css
from ...assets import load_js
from ...table import TABLE_INDEX_KEY
from ...table import prepare_table_data
from ...util import sanitize_html
from ..icons import Icon

logger = logging.getLogger(__name__)

DEFAULT_ID_WIDTH = 110
env = jinja2.Environment(loader=jinja2.PackageLoader("syft", "assets/jinja"))  # nosec


def create_tabulator_columns(
    column_names: list[str],
    column_widths: dict | None = None,
    header_sort: bool = True,
) -> tuple[list[dict], dict | None]:
    """Returns tuple of (columns, row_header) for tabulator table"""
    if column_widths is None:
        column_widths = {}

    columns = []
    row_header = {}
    if TABLE_INDEX_KEY in column_names:
        row_header = {
            "field": TABLE_INDEX_KEY,
            "frozen": True,
            "widthGrow": 0.3,
            "minWidth": 60,
            "headerSort": header_sort,
        }

    for colname in column_names:
        if colname != TABLE_INDEX_KEY:
            column = {
                "title": colname,
                "field": colname,
                "formatter": "html",
                "resizable": True,
                "minWidth": 60,
                "maxInitialWidth": 500,
                "headerSort": header_sort,
            }
            if colname in column_widths:
                column["widthGrow"] = column_widths[colname]
            columns.append(column)

    return columns, row_header


def format_dict(data: Any) -> str:
    # relative
    from .sync import Badge
    from .sync import CopyButton
    from .sync import Label

    if not isinstance(data, dict):
        return data

    if set(data.keys()) != {"type", "value"}:
        return sanitize_html(str(data))
    if "badge" in data["type"]:
        return Badge(value=data["value"], badge_class=data["type"]).to_html()
    elif "label" in data["type"]:
        return Label(value=data["value"], label_class=data["type"]).to_html()
    if "clipboard" in data["type"]:
        return CopyButton(copy_text=data["value"]).to_html()

    return sanitize_html(str(data))


def format_uid(uid: UID) -> str:
    # relative
    from .sync import CopyButton

    return CopyButton(copy_text=uid.no_dash).to_html()


def format_table_data(table_data: list[dict[str, Any]]) -> list[dict[str, str]]:
    formatted: list[dict[str, str]] = []
    for row in table_data:
        row_formatted: dict[str, str] = {}
        for k, v in row.items():
            if isinstance(v, str):
                row_formatted[k] = sanitize_html(v.replace("\n", "<br>"))
                continue
            # make UID copyable and trimmed
            if isinstance(v, UID):
                v_formatted = format_uid(v)
            else:
                v_formatted = format_dict(v)
            row_formatted[k] = v_formatted
        formatted.append(row_formatted)
    return formatted


def _render_tabulator_table(
    uid: str,
    table_data: list[dict],
    table_metadata: dict,
    max_height: int | None,
    pagination: bool,
    header_sort: bool,
) -> str:
    table_template = env.get_template("table.jinja2")
    tabulator_js = load_js("tabulator.min.js")
    tabulator_css = load_css("tabulator_pysyft.min.css")
    js = load_js("table.js")
    css = load_css("style.css")

    # Add tabulator as a named module for VSCode compatibility
    tabulator_js = tabulator_js.replace(
        "define(t)", "define('tabulator-tables', [], t)"
    )

    icon = table_metadata.get("icon", None)
    if icon is None:
        icon = Icon.TABLE.svg

    column_data, row_header = create_tabulator_columns(
        table_metadata["columns"], header_sort=header_sort
    )
    table_data = format_table_data(table_data)
    table_html = table_template.render(
        uid=uid,
        columns=json.dumps(column_data),
        row_header=json.dumps(row_header),
        data=json.dumps(table_data),
        css=css,
        js=js,
        index_field_name=TABLE_INDEX_KEY,
        icon=icon,
        name=table_metadata["name"],
        tabulator_js=tabulator_js,
        tabulator_css=tabulator_css,
        max_height=json.dumps(max_height),
        pagination=json.dumps(pagination),
        header_sort=json.dumps(header_sort),
    )

    return table_html


def build_tabulator_table_with_data(
    table_data: list[dict],
    table_metadata: dict,
    uid: str | None = None,
    max_height: int | None = None,
    pagination: bool = True,
    header_sort: bool = True,
) -> str:
    """
    Builds a Tabulator table for the provided data and metadata.

    Args:
        table_data (list[dict]): The data to populate the table.
        table_metadata (dict): The metadata for the table.
        uid (str, optional): The unique identifier for the table. Defaults to None.
        max_height (int, optional): The maximum height of the table. Defaults to None.
        pagination (bool, optional): Whether to enable pagination. Defaults to True.
        header_sort (bool, optional): Whether to enable header sorting. Defaults to True.

    Returns:
        str: The HTML representation of the Tabulator table.

    """
    uid = uid if uid is not None else secrets.token_hex(4)
    return _render_tabulator_table(
        uid, table_data, table_metadata, max_height, pagination, header_sort
    )


def build_tabulator_table(
    obj: Any,
    uid: str | None = None,
    max_height: int | None = None,
    pagination: bool = True,
    header_sort: bool = True,
) -> str | None:
    """
    Builds a Tabulator table from the given object if possible.

    If the object cannot be represented as a table, returns None.

    Args:
        obj (Any): The object to build the table from.
        uid (str, optional): The unique identifier for the table. Defaults to None.
        max_height (int, optional): The maximum height of the table. Defaults to None.
        pagination (bool, optional): Whether to enable pagination. Defaults to True.
        header_sort (bool, optional): Whether to enable header sorting. Defaults to True.

    Returns:
        str | None: The HTML representation of the Tabulator table or None

    """
    table_data, table_metadata = prepare_table_data(obj)
    if len(table_data) == 0:
        if hasattr(obj, "__len__") and len(obj) == 0:
            return obj.__repr__()
        else:
            return None

    return build_tabulator_table_with_data(
        table_data, table_metadata, uid, max_height, pagination, header_sort
    )


def show_table(obj: Any) -> None:
    table = build_tabulator_table(obj)
    if table is not None:
        display(HTML(table))


def highlight_single_row(
    table_uid: str,
    index: int | str | None = None,
    jump_to_row: bool = True,
) -> None:
    js_code = f"<script>highlightSingleRow('{table_uid}', {json.dumps(index)}, {json.dumps(jump_to_row)});</script>"
    display(HTML(js_code))


def update_table_cell(uid: str, index: int, field: str, value: str) -> None:
    js_code = f"""
    <script>
    updateTableCell('{uid}', {json.dumps(index)}, {json.dumps(field)}, {json.dumps(value)});
    </script>
    """
    display(HTML(js_code))



# stdlib
from collections.abc import Callable
import operator
from typing import Any

# third party
from packaging import version

operators = {
    ">": operator.gt,
    ">=": operator.ge,
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
}


def get_operator(version_string: str) -> tuple[str, Callable, str]:
    op: Any = operator.ge
    op_char: str = ">="
    if len(version_string) > 2:
        two_char = version_string[0:2]
        one_char = version_string[0:1]
        if two_char in operators.keys():
            op_char = two_char
            version_string = version_string[2:]
        elif one_char in operators.keys():
            op_char = one_char
            version_string = version_string[1:]
    op = operators[op_char]
    return version_string, op, op_char


def check_rule(
    version_string: str, LATEST_STABLE_SYFT: str, __version__: str
) -> tuple[Any, list[str], list[str]]:
    version_string, op, op_char = get_operator(version_string)
    syft_version = version.parse(__version__)
    stable_version = version.parse(LATEST_STABLE_SYFT)
    required = version.parse(version_string)
    result = op(syft_version, required)

    requirements = []
    messages = []

    if result:
        requirements.append(f"the requirement {op_char}{required}")
    else:
        requirements.append(f"the requirement {op_char}{required}")
        pre = ""
        if required.minor > stable_version.minor:
            pre = " --pre"
        msg = f"Alternatively you could try to match {op_char}{required} with:\n"
        if required > syft_version:
            upgrade = f"pip install -U{pre} syft or "
        else:
            upgrade = ""
        msg += f"{upgrade}pip install syft=={required}"
        messages.append(msg)
    return result, requirements, messages


def make_requires(LATEST_STABLE_SYFT: str, __version__: str) -> Callable:
    def requires(version_string: str, silent: bool = False) -> bool | None:
        syft_version = version.parse(__version__)
        parts = version_string.split(",")
        result = True
        all_requirements = []
        all_messages = []
        for part in parts:
            part_result, requirements, messages = check_rule(
                version_string=part,
                LATEST_STABLE_SYFT=LATEST_STABLE_SYFT,
                __version__=__version__,
            )
            all_requirements += requirements
            all_messages += messages
            if not part_result:
                result = False

        if silent:
            return result

        msg_requirements = " and ".join(all_requirements)
        if result:
            print(
                f"✅ The installed version of syft=={syft_version} matches {msg_requirements}"
            )
        else:
            print(
                f"❌ The installed version of syft=={syft_version} doesn't match {msg_requirements}"
            )
        if len(all_messages):
            print("This code or notebook may have issues if APIs have changed\n")
            print("\n\n".join(all_messages))
        return None

    return requires



# stdlib

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..code.user_code import SubmitUserCode
from ..code.user_code import UserCode
from ..code.user_code_service import UserCodeService
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .code_history import CodeHistoriesDict
from .code_history import CodeHistory
from .code_history import CodeHistoryView
from .code_history import UsersCodeHistoriesDict
from .code_history_stash import CodeHistoryStash


@instrument
@serializable(canonical_name="CodeHistoryService", version=1)
class CodeHistoryService(AbstractService):
    store: DocumentStore
    stash: CodeHistoryStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = CodeHistoryStash(store=store)

    @service_method(
        path="code_history.submit_version",
        name="submit_version",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def submit_version(
        self,
        context: AuthedServiceContext,
        code: SubmitUserCode | UserCode,
        comment: str | None = None,
    ) -> SyftSuccess | SyftError:
        user_code_service = context.server.get_service("usercodeservice")
        if isinstance(code, SubmitUserCode):
            result = user_code_service._submit(context=context, code=code)
            if result.is_err():
                return SyftError(message=str(result.err()))
            code = result.ok()

        result = self.stash.get_by_service_func_name_and_verify_key(
            credentials=context.credentials,
            service_func_name=code.service_func_name,
            user_verify_key=context.credentials,
        )

        if result.is_err():
            return SyftError(message=result.err())

        code_history: CodeHistory | None = result.ok()

        if code_history is None:
            code_history = CodeHistory(
                id=UID(),
                server_uid=context.server.id,
                user_verify_key=context.credentials,
                service_func_name=code.service_func_name,
            )
            result = self.stash.set(credentials=context.credentials, obj=code_history)
            if result.is_err():
                return SyftError(message=result.err())

        code_history.add_code(code=code, comment=comment)
        result = self.stash.update(credentials=context.credentials, obj=code_history)
        if result.is_err():
            return SyftError(message=result.err())

        return SyftSuccess(message="Code version submit success")

    @service_method(
        path="code_history.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_all(self, context: AuthedServiceContext) -> list[CodeHistory] | SyftError:
        """Get a Dataset"""
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="code_history.get", name="get", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_code_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        """Get a User Code Item"""
        result = self.stash.get_by_uid(context.credentials, uid=uid)
        if result.is_ok():
            code_history = result.ok()
            return code_history
        return SyftError(message=result.err())

    @service_method(path="code_history.delete", name="delete")
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_ok():
            return result.ok()
        else:
            return SyftError(message=result.err())

    def fetch_histories_for_user(
        self, context: AuthedServiceContext, user_verify_key: SyftVerifyKey
    ) -> CodeHistoriesDict | SyftError:
        if context.role in [ServiceRole.DATA_OWNER, ServiceRole.ADMIN]:
            result = self.stash.get_by_verify_key(
                credentials=context.server.verify_key, user_verify_key=user_verify_key
            )
        else:
            result = self.stash.get_by_verify_key(
                credentials=context.credentials, user_verify_key=user_verify_key
            )

        user_code_service: UserCodeService = context.server.get_service(
            "usercodeservice"
        )  # type: ignore

        def get_code(uid: UID) -> UserCode | SyftError:
            return user_code_service.stash.get_by_uid(
                credentials=context.server.verify_key,
                uid=uid,
            ).ok()

        if result.is_ok():
            code_histories = result.ok()
            code_versions_dict = {}

            for code_history in code_histories:
                user_code_list = [
                    get_code(uid) for uid in code_history.user_code_history
                ]
                code_versions = CodeHistoryView(
                    user_code_history=user_code_list,
                    service_func_name=code_history.service_func_name,
                    comment_history=code_history.comment_history,
                )
                code_versions_dict[code_history.service_func_name] = code_versions
            return CodeHistoriesDict(code_versions=code_versions_dict)
        else:
            return SyftError(message=result.err())

    @service_method(
        path="code_history.get_history",
        name="get_history",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_histories_for_current_user(
        self, context: AuthedServiceContext
    ) -> CodeHistoriesDict | SyftError:
        return self.fetch_histories_for_user(
            context=context, user_verify_key=context.credentials
        )

    @service_method(
        path="code_history.get_history_for_user",
        name="get_history_for_user",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_history_for_user(
        self, context: AuthedServiceContext, email: str
    ) -> CodeHistoriesDict | SyftError:
        user_service = context.server.get_service("userservice")
        result = user_service.stash.get_by_email(
            credentials=context.credentials, email=email
        )
        if result.is_ok():
            user = result.ok()
            return self.fetch_histories_for_user(
                context=context, user_verify_key=user.verify_key
            )
        return SyftError(message=result.err())

    @service_method(
        path="code_history.get_histories",
        name="get_histories",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_histories_group_by_user(
        self, context: AuthedServiceContext
    ) -> UsersCodeHistoriesDict | SyftError:
        if context.role in [ServiceRole.DATA_OWNER, ServiceRole.ADMIN]:
            result = self.stash.get_all(context.credentials, has_permission=True)
        else:
            result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=result.err())
        code_histories: list[CodeHistory] = result.ok()

        user_service = context.server.get_service("userservice")
        result = user_service.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=result.err())
        users = result.ok()

        user_code_histories = UsersCodeHistoriesDict(server_uid=context.server.id)

        verify_key_2_user_email = {}
        for user in users:
            user_code_histories.user_dict[user.email] = []
            verify_key_2_user_email[user.verify_key] = user.email

        for code_history in code_histories:
            user_email = verify_key_2_user_email[code_history.user_verify_key]
            user_code_histories.user_dict[user_email].append(
                code_history.service_func_name
            )

        return user_code_histories

    @service_method(
        path="code_history.get_by_name_and_user_email",
        name="get_by_name_and_user_email",
    )
    def get_by_func_name_and_user_email(
        self,
        context: AuthedServiceContext,
        service_func_name: str,
        user_email: str,
        user_id: UID,
    ) -> list[CodeHistory] | SyftError:
        user_service = context.server.get_service("userservice")
        user_verify_key = user_service.user_verify_key(user_email)

        if isinstance(user_verify_key, SyftError):
            return user_verify_key

        kwargs = {
            "id": user_id,
            "email": user_email,
            "verify_key": user_verify_key,
            "service_func_name": service_func_name,
        }

        result = self.stash.find_all(credentials=context.credentials, **kwargs)
        if result.is_err():  # or len(result) > 1
            return result

        return result.ok()



# stdlib
import argparse

# relative
from ..orchestra import Orchestra
from ..orchestra import ServerHandle


def str_to_bool(bool_str: str | None) -> bool:
    result = False
    bool_str = str(bool_str).lower()
    if bool_str == "true" or bool_str == "1":
        result = True
    return result


def run() -> ServerHandle | None:
    parser = argparse.ArgumentParser()
    parser.add_argument("command", help="command: launch", type=str, default="none")
    parser.add_argument(
        "--name", help="server name", type=str, default="syft-server", dest="name"
    )
    parser.add_argument(
        "--server-type",
        help="server type",
        type=str,
        default="python",
        dest="server_type",
    )
    parser.add_argument(
        "--host",
        help="host for binding",
        type=str,
        default="0.0.0.0",  # nosec
        dest="host",
    )

    parser.add_argument(
        "--port", help="port for binding", type=int, default=8080, dest="port"
    )
    parser.add_argument(
        "--dev-mode",
        help="developer mode",
        type=str,
        default="True",
        dest="dev_mode",
    )
    parser.add_argument(
        "--reset",
        help="reset",
        type=str,
        default="True",
        dest="reset",
    )
    parser.add_argument(
        "--local-db",
        help="reset",
        type=str,
        default="False",
        dest="local_db",
    )
    parser.add_argument(
        "--processes",
        help="processing mode",
        type=int,
        default=0,
        dest="processes",
    )
    parser.add_argument(
        "--tail",
        help="tail mode",
        type=str,
        default="True",
        dest="tail",
    )

    args = parser.parse_args()
    if args.command != "launch":
        print("syft launch is the only command currently supported")

    args.dev_mode = str_to_bool(args.dev_mode)
    args.reset = str_to_bool(args.reset)
    args.local_db = str_to_bool(args.local_db)
    args.tail = str_to_bool(args.tail)
    args.cmd = str_to_bool(args.cmd)

    server = Orchestra.launch(
        name=args.name,
        server_type=args.server_type,
        host=args.host,
        port=args.port,
        dev_mode=args.dev_mode,
        reset=args.reset,
        local_db=args.local_db,
        processes=args.processes,
        tail=args.tail,
    )
    if not args.tail:
        return server
    return None



# stdlib
from collections.abc import Callable
from enum import Enum
import logging

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...abstract_server import ServerType
from ...client.client import ServerConnection
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...service.response import SyftError
from ...types.datetime import DateTime
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.transforms import TransformContext
from ...types.uid import UID
from ..context import ServerServiceContext
from ..metadata.server_metadata import ServerMetadata
from .routes import HTTPServerRoute
from .routes import PythonServerRoute
from .routes import ServerRoute
from .routes import ServerRouteType
from .routes import VeilidServerRoute
from .routes import connection_to_route
from .routes import route_to_connection

logger = logging.getLogger(__name__)


@serializable(canonical_name="ServerPeerConnectionStatus", version=1)
class ServerPeerConnectionStatus(Enum):
    ACTIVE = "ACTIVE"
    INACTIVE = "INACTIVE"
    TIMEOUT = "TIMEOUT"


@serializable()
class ServerPeer(SyftObject):
    # version
    __canonical_name__ = "ServerPeer"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_searchable__ = ["name", "server_type"]
    __attr_unique__ = ["verify_key"]
    __repr_attrs__ = [
        "name",
        "server_type",
        "admin_email",
        "ping_status",
        "ping_status_message",
        "pinged_timestamp",
    ]

    id: UID | None = None  # type: ignore[assignment]
    name: str
    verify_key: SyftVerifyKey
    server_routes: list[ServerRouteType] = []
    server_type: ServerType
    admin_email: str
    ping_status: ServerPeerConnectionStatus | None = None
    ping_status_message: str | None = None
    pinged_timestamp: DateTime | None = None

    def existed_route(self, route: ServerRouteType) -> tuple[bool, int | None]:
        """Check if a route exists in self.server_routes

        Args:
            route: the route to be checked. For now it can be either
                HTTPServerRoute or PythonServerRoute

        Returns:
            if the route exists, returns (True, index of the existed route in self.server_routes)
            if the route does not exist returns (False, None)
        """

        if route:
            if not isinstance(
                route, HTTPServerRoute | PythonServerRoute | VeilidServerRoute
            ):
                raise ValueError(f"Unsupported route type: {type(route)}")
            for i, r in enumerate(self.server_routes):
                if route == r:
                    return (True, i)

        return (False, None)

    def update_route_priority(self, route: ServerRoute) -> ServerRoute:
        """
        Assign the new_route's priority to be current max + 1

        Args:
            route (ServerRoute): The new route whose priority is to be updated.

        Returns:
            ServerRoute: The new route with the updated priority
        """
        current_max_priority: int = max(route.priority for route in self.server_routes)
        route.priority = current_max_priority + 1
        return route

    def pick_highest_priority_route(self, oldest: bool = True) -> ServerRoute:
        """
        Picks the route with the highest priority from the list of server routes.

        Args:
            oldest (bool):
                If True, picks the oldest route to have the highest priority,
                    meaning the route with min priority value.
                If False, picks the most recent route with the highest priority,
                    meaning the route with max priority value.

        Returns:
            ServerRoute: The route with the highest priority.

        """
        highest_priority_route: ServerRoute = self.server_routes[-1]
        for route in self.server_routes[:-1]:
            if oldest:
                if route.priority < highest_priority_route.priority:
                    highest_priority_route = route
            else:
                if route.priority > highest_priority_route.priority:
                    highest_priority_route = route
        return highest_priority_route

    def update_route(self, route: ServerRoute) -> None:
        """
        Update the route for the server.
        If the route already exists, return it.
        If the route is new, assign it to have the priority of (current_max + 1)

        Args:
            route (ServerRoute): The new route to be added to the peer.
        """
        existed, idx = self.existed_route(route)
        if existed:
            self.server_routes[idx] = route  # type: ignore
        else:
            new_route = self.update_route_priority(route)
            self.server_routes.append(new_route)

    def update_routes(self, new_routes: list[ServerRoute]) -> None:
        """
        Update multiple routes of the server peer.

        This method takes a list of new routes as input.
        It first updates the priorities of the new routes.
        Then, for each new route, it checks if the route already exists for the server peer.
        If it does, it updates the priority of the existing route.
        If it doesn't, it adds the new route to the server.

        Args:
            new_routes (list[ServerRoute]): The new routes to be added to the server.

        Returns:
            None
        """
        for new_route in new_routes:
            self.update_route(new_route)

    def update_existed_route_priority(
        self, route: ServerRoute, priority: int | None = None
    ) -> ServerRouteType | SyftError:
        """
        Update the priority of an existed route.

        Args:
            route (ServerRoute): The route whose priority is to be updated.
            priority (int | None): The new priority of the route. If not given,
                the route will be assigned with the highest priority.

        Returns:
            ServerRoute: The route with updated priority if the route exists
            SyftError: If the route does not exist or the priority is invalid
        """
        if priority is not None and priority <= 0:
            return SyftError(
                message="Priority must be greater than 0. Now it is {priority}."
            )

        existed, index = self.existed_route(route=route)

        if not existed or index is None:
            return SyftError(message=f"Route with id {route.id} does not exist.")

        if priority is not None:
            self.server_routes[index].priority = priority
        else:
            self.server_routes[index].priority = self.update_route_priority(
                route
            ).priority

        return self.server_routes[index]

    @staticmethod
    def from_client(client: SyftClient) -> "ServerPeer":
        if not client.metadata:
            raise ValueError("Client has to have metadata first")

        peer = client.metadata.to(ServerMetadata).to(ServerPeer)
        route = connection_to_route(client.connection)
        peer.server_routes.append(route)
        return peer

    @property
    def latest_added_route(self) -> ServerRoute | None:
        """
        Returns the latest added route from the list of server routes.

        Returns:
            ServerRoute | None: The latest added route, or None if there are no routes.
        """
        return self.server_routes[-1] if self.server_routes else None

    def client_with_context(
        self, context: ServerServiceContext
    ) -> Result[type[SyftClient], str]:
        # third party

        if len(self.server_routes) < 1:
            raise ValueError(f"No routes to peer: {self}")
        # select the route with highest priority to connect to the peer
        final_route: ServerRoute = self.pick_highest_priority_route()
        connection: ServerConnection = route_to_connection(route=final_route)
        try:
            client_type = connection.get_client_type()
        except Exception as e:
            msg = f"Failed to establish a connection with {self.server_type} '{self.name}'"
            logger.error(msg, exc_info=e)
            return Err(msg)
        if isinstance(client_type, SyftError):
            return Err(client_type.message)
        return Ok(
            client_type(connection=connection, credentials=context.server.signing_key)
        )

    def client_with_key(self, credentials: SyftSigningKey) -> SyftClient | SyftError:
        if len(self.server_routes) < 1:
            raise ValueError(f"No routes to peer: {self}")

        final_route: ServerRoute = self.pick_highest_priority_route()

        connection = route_to_connection(route=final_route)
        client_type = connection.get_client_type()
        if isinstance(client_type, SyftError):
            return client_type

        return client_type(connection=connection, credentials=credentials)

    @property
    def guest_client(self) -> SyftClient:
        guest_key = SyftSigningKey.generate()
        return self.client_with_key(credentials=guest_key)

    def proxy_from(self, client: SyftClient) -> SyftClient:
        return client.proxy_to(self)

    def get_rtunnel_route(self) -> HTTPServerRoute | None:
        for route in self.server_routes:
            if hasattr(route, "rtunnel_token") and route.rtunnel_token:
                return route
        return None

    def delete_route(self, route: ServerRouteType) -> SyftError | None:
        """
        Deletes a route from the peer's route list.
        Takes O(n) where is n is the number of routes in self.server_routes.

        Args:
            route (ServerRouteType): The route to be deleted;

        Returns:
            SyftError: If failing to delete server route
        """
        if route:
            try:
                self.server_routes = [r for r in self.server_routes if r != route]
            except Exception as e:
                return SyftError(
                    message=f"Error deleting route with id {route.id}. Exception: {e}"
                )

        return None


@serializable()
class ServerPeerUpdate(PartialSyftObject):
    __canonical_name__ = "ServerPeerUpdate"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    name: str
    server_routes: list[ServerRouteType]
    admin_email: str
    ping_status: ServerPeerConnectionStatus
    ping_status_message: str
    pinged_timestamp: DateTime


def drop_veilid_route() -> Callable:
    def _drop_veilid_route(context: TransformContext) -> TransformContext:
        if context.output:
            server_routes = context.output["server_routes"]
            new_routes = [
                server_route
                for server_route in server_routes
                if not isinstance(server_route, VeilidServerRoute)
            ]
            context.output["server_routes"] = new_routes
        return context

    return _drop_veilid_route



# stdlib

# third party
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...types.datetime import DateTime
from ...types.uid import UID
from ...util.telemetry import instrument
from .request import Request

RequestingUserVerifyKeyPartitionKey = PartitionKey(
    key="requesting_user_verify_key", type_=SyftVerifyKey
)

OrderByRequestTimeStampPartitionKey = PartitionKey(key="request_time", type_=DateTime)


@instrument
@serializable(canonical_name="RequestStash", version=1)
class RequestStash(BaseUIDStoreStash):
    object_type = Request
    settings: PartitionSettings = PartitionSettings(
        name=Request.__canonical_name__, object_type=Request
    )

    def get_all_for_verify_key(
        self,
        credentials: SyftVerifyKey,
        verify_key: SyftVerifyKey,
    ) -> Result[list[Request], str]:
        if isinstance(verify_key, str):
            verify_key = SyftVerifyKey.from_string(verify_key)
        qks = QueryKeys(qks=[RequestingUserVerifyKeyPartitionKey.with_obj(verify_key)])
        return self.query_all(
            credentials=credentials,
            qks=qks,
            order_by=OrderByRequestTimeStampPartitionKey,
        )

    def get_by_usercode_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> Result[list[Request], str]:
        query = self.get_all(credentials=credentials)
        if query.is_err():
            return query

        all_requests: list[Request] = query.ok()
        results = [r for r in all_requests if r.code_id == user_code_id]
        return Ok(results)



# future
from __future__ import annotations

# stdlib
import logging
from pathlib import Path
import re
from string import Template
import traceback
from typing import TYPE_CHECKING
from typing import cast

# third party
import markdown
from result import Result
from tqdm import tqdm

# relative
from ..abstract_server import ServerSideType
from ..serde.serializable import serializable
from ..service.action.action_object import ActionObject
from ..service.code_history.code_history import CodeHistoriesDict
from ..service.code_history.code_history import UsersCodeHistoriesDict
from ..service.dataset.dataset import Contributor
from ..service.dataset.dataset import CreateAsset
from ..service.dataset.dataset import CreateDataset
from ..service.migration.object_migration_state import MigrationData
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..service.response import SyftWarning
from ..service.sync.diff_state import ResolvedSyncState
from ..service.sync.sync_state import SyncState
from ..service.user.roles import Roles
from ..service.user.user import UserView
from ..types.blob_storage import BlobFile
from ..types.uid import UID
from ..util.misc_objs import HTMLObject
from ..util.util import get_mb_size
from ..util.util import prompt_warning_message
from .api import APIModule
from .client import SyftClient
from .client import login
from .client import login_as_guest
from .connection import ServerConnection
from .protocol import SyftProtocol

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # relative
    from ..orchestra import ServerHandle
    from ..service.project.project import Project


def _get_files_from_glob(glob_path: str) -> list[Path]:
    files = Path().glob(glob_path)
    return [f for f in files if f.is_file() and not f.name.startswith(".")]


def _get_files_from_dir(dir: Path, recursive: bool) -> list:
    files = dir.rglob("*") if recursive else dir.iterdir()
    return [f for f in files if not f.name.startswith(".") and f.is_file()]


def _contains_subdir(dir: Path) -> bool:
    for item in dir.iterdir():
        if item.is_dir():
            return True
    return False


def add_default_uploader(
    user: UserView, obj: CreateDataset | CreateAsset
) -> CreateDataset | CreateAsset:
    uploader = None
    for contributor in obj.contributors:
        if contributor.role == str(Roles.UPLOADER):
            uploader = contributor
            break

    if uploader is None:
        uploader = Contributor(
            role=str(Roles.UPLOADER),
            name=user.name,
            email=user.email,
        )
        obj.contributors.add(uploader)

    obj.uploader = uploader
    return obj


@serializable(canonical_name="DatasiteClient", version=1)
class DatasiteClient(SyftClient):
    def __repr__(self) -> str:
        return f"<DatasiteClient: {self.name}>"

    def upload_dataset(self, dataset: CreateDataset) -> SyftSuccess | SyftError:
        # relative
        from ..types.twin_object import TwinObject

        if self.users is None:
            return SyftError(f"can't get user service for {self}")

        user = self.users.get_current_user()
        dataset = add_default_uploader(user, dataset)
        for i in range(len(dataset.asset_list)):
            asset = dataset.asset_list[i]
            dataset.asset_list[i] = add_default_uploader(user, asset)

        # dataset._check_asset_must_contain_mock()
        dataset_size: float = 0.0

        # TODO: Refactor so that object can also be passed to generate warnings

        self.api.connection = cast(ServerConnection, self.api.connection)

        metadata = self.api.connection.get_server_metadata(self.api.signing_key)

        if (
            metadata.show_warnings
            and metadata.server_side_type == ServerSideType.HIGH_SIDE.value
        ):
            message = (
                "You're approving a request on "
                f"{metadata.server_side_type} side {metadata.server_type} "
                "which may host datasets with private information."
            )
            prompt_warning_message(message=message, confirm=True)

        with tqdm(
            total=len(dataset.asset_list), colour="green", desc="Uploading"
        ) as pbar:
            for asset in dataset.asset_list:
                try:
                    contains_empty: bool = asset.contains_empty()
                    twin = TwinObject(
                        private_obj=ActionObject.from_obj(asset.data),
                        mock_obj=ActionObject.from_obj(asset.mock),
                        syft_server_location=self.id,
                        syft_client_verify_key=self.verify_key,
                    )
                    res = twin._save_to_blob_storage(allow_empty=contains_empty)
                    if isinstance(res, SyftError):
                        return res
                except Exception as e:
                    tqdm.write(f"Failed to create twin for {asset.name}. {e}")
                    return SyftError(message=f"Failed to create twin. {e}")

                if isinstance(res, SyftWarning):
                    logger.debug(res.message)
                response = self.api.services.action.set(
                    twin, ignore_detached_objs=contains_empty
                )
                if isinstance(response, SyftError):
                    tqdm.write(f"Failed to upload asset: {asset.name}")
                    return response

                asset.action_id = twin.id
                asset.server_uid = self.id
                dataset_size += get_mb_size(asset.data)

                # Update the progress bar and set the dynamic description
                pbar.set_description(f"Uploading: {asset.name}")
                pbar.update(1)

        dataset.mb_size = dataset_size
        valid = dataset.check()
        if isinstance(valid, SyftError):
            return valid
        return self.api.services.dataset.add(dataset=dataset)

    def refresh(self) -> None:
        if self.credentials:
            self._fetch_server_metadata(self.credentials)

        if self._api and self._api.refresh_api_callback:
            self._api.refresh_api_callback()

    def get_sync_state(self) -> SyncState | SyftError:
        state: SyncState = self.api.services.sync._get_state()
        if isinstance(state, SyftError):
            return state

        for uid, obj in state.objects.items():
            if isinstance(obj, ActionObject):
                obj = obj.refresh_object(resolve_nested=False)
                state.objects[uid] = obj
        return state

    def apply_state(self, resolved_state: ResolvedSyncState) -> SyftSuccess | SyftError:
        if len(resolved_state.delete_objs):
            raise NotImplementedError("TODO implement delete")
        items = resolved_state.create_objs + resolved_state.update_objs

        action_objects = [x for x in items if isinstance(x, ActionObject)]

        for action_object in action_objects:
            action_object.reload_cache()
            # NOTE permissions are added separately server side
            action_object._send(self.id, self.verify_key, add_storage_permission=False)
            action_object._clear_cache()

        ignored_batches = resolved_state.ignored_batches

        res = self.api.services.sync.sync_items(
            items,
            resolved_state.new_permissions,
            resolved_state.new_storage_permissions,
            ignored_batches,
            unignored_batches=resolved_state.unignored_batches,
        )
        if isinstance(res, SyftError):
            return res

        self._fetch_api(self.credentials)
        return res

    def upload_files(
        self,
        file_list: BlobFile | list[BlobFile] | str | list[str] | Path | list[Path],
        allow_recursive: bool = False,
        show_files: bool = False,
    ) -> SyftSuccess | SyftError:
        if not file_list:
            return SyftError(message="No files to upload")

        if not isinstance(file_list, list):
            file_list = [file_list]  # type: ignore[assignment]
        file_list = cast(list, file_list)

        expanded_file_list: list[BlobFile | Path] = []

        for file in file_list:
            if isinstance(file, BlobFile):
                expanded_file_list.append(file)
                continue

            path = Path(file)

            if re.search(r"[\*\?\[]", str(path)):
                expanded_file_list.extend(_get_files_from_glob(str(path)))
            elif path.is_dir():
                if not allow_recursive and _contains_subdir(path):
                    res = input(
                        f"Do you want to include all files recursively in {path.absolute()}? [y/n]: "
                    ).lower()
                    print(
                        f'{"Recursively uploading all files" if res == "y" else "Uploading files"} in {path.absolute()}'
                    )
                    allow_recursive = res == "y"
                expanded_file_list.extend(_get_files_from_dir(path, allow_recursive))
            elif path.exists():
                expanded_file_list.append(path)

        if not expanded_file_list:
            return SyftError(message="No files to upload were found")

        print(
            f"Uploading {len(expanded_file_list)} {'file' if len(expanded_file_list) == 1 else 'files'}:"
        )

        if show_files:
            for file in expanded_file_list:
                if isinstance(file, BlobFile):
                    print(file.path or file.file_name)
                else:
                    print(file.absolute())

        try:
            result = []
            for file in expanded_file_list:
                if not isinstance(file, BlobFile):
                    file = BlobFile(path=file, file_name=file.name)
                print("Uploading", file.file_name)
                if not file.uploaded:
                    file.upload_to_blobstorage(self)
                result.append(file)

            return ActionObject.from_obj(result).send(self)
        except Exception as err:
            return SyftError(
                message=f"Failed to upload files: {err}.\n{traceback.format_exc()}"
            )

    def connect_to_gateway(
        self,
        via_client: SyftClient | None = None,
        url: str | None = None,
        port: int | None = None,
        handle: ServerHandle | None = None,  # noqa: F821
        email: str | None = None,
        password: str | None = None,
        protocol: str | SyftProtocol = SyftProtocol.HTTP,
        reverse_tunnel: bool = False,
    ) -> SyftSuccess | SyftError | None:
        if isinstance(protocol, str):
            protocol = SyftProtocol(protocol)

        if via_client is not None:
            client = via_client
        elif handle is not None:
            client = handle.client
        else:
            client = (
                login_as_guest(url=url, port=port)
                if email is None
                else login(url=url, port=port, email=email, password=password)
            )
            if isinstance(client, SyftError):
                return client

        res = self.exchange_route(
            client,
            protocol=protocol,
            reverse_tunnel=reverse_tunnel,
        )
        if isinstance(res, SyftSuccess):
            if self.metadata:
                return SyftSuccess(
                    message=(
                        f"Connected {self.metadata.server_type} "
                        f"'{self.metadata.name}' to gateway '{client.name}'. "
                        f"{res.message}"
                    )
                )
            else:
                return SyftSuccess(message=f"Connected to '{client.name}' gateway")

        return res

    def _get_service_by_name_if_exists(self, name: str) -> APIModule | None:
        if self.api.has_service(name):
            return getattr(self.api.services, name)
        return None

    def set_server_side_type_dangerous(
        self, server_side_type: str
    ) -> Result[SyftSuccess, SyftError]:
        return self.api.services.settings.set_server_side_type_dangerous(
            server_side_type
        )

    @property
    def data_subject_registry(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("data_subject")

    @property
    def code(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("code")

    @property
    def worker(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("worker")

    @property
    def requests(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("request")

    @property
    def datasets(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("dataset")

    @property
    def projects(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("project")

    @property
    def code_history_service(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("code_history")

    @property
    def code_history(self) -> CodeHistoriesDict:
        return self.api.services.code_history.get_history()

    @property
    def code_histories(self) -> UsersCodeHistoriesDict:
        return self.api.services.code_history.get_histories()

    @property
    def images(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("worker_image")

    @property
    def worker_pools(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("worker_pool")

    @property
    def worker_images(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("worker_images")

    @property
    def sync(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("sync")

    @property
    def code_status(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("code_status")

    @property
    def output(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("output")

    @property
    def migration(self) -> APIModule | None:
        return self._get_service_by_name_if_exists("migration")

    def get_migration_data(
        self, include_blobs: bool = True
    ) -> MigrationData | SyftError:
        res = self.api.services.migration.get_migration_data()
        if isinstance(res, SyftError):
            return res

        if include_blobs:
            res.download_blobs()

        return res

    def load_migration_data(self, path: str | Path) -> SyftSuccess | SyftError:
        migration_data = MigrationData.from_file(path)
        if isinstance(migration_data, SyftError):
            return migration_data
        migration_data._set_obj_location_(self.id, self.verify_key)

        if self.id != migration_data.server_uid:
            return SyftError(
                message=f"This Migration data is not for this server. Expected server id {self.id}, "
                f"got {migration_data.server_uid}"
            )

        if migration_data.signing_key.verify_key != self.verify_key:
            return SyftError(
                message="Root verify key in migration data does not match this client's verify key"
            )

        res = migration_data.migrate_and_upload_blobs()
        if isinstance(res, SyftError):
            return res

        migration_data_without_blobs = migration_data.copy_without_blobs()
        return self.api.services.migration.apply_migration_data(
            migration_data_without_blobs
        )

    def get_project(
        self,
        name: str | None = None,
        uid: UID | None = None,
    ) -> Project | None:
        """Get project by name or UID"""

        if not self.api.has_service("project"):
            return None

        if name:
            return self.api.services.project.get_by_name(name)

        elif uid:
            return self.api.services.project.get_by_uid(uid)

        return self.api.services.project.get_all()

    def _repr_html_(self) -> str:
        obj = self.api.services.settings.welcome_show()
        if isinstance(obj, SyftError):
            return obj.message
        updated_template_str = Template(obj.text).safe_substitute(
            server_url=getattr(self.connection, "url", None)
        )
        # If it's a markdown structured file
        if not isinstance(obj, HTMLObject):
            return markdown.markdown(updated_template_str)

        # if it's a html string
        return updated_template_str
